"""An image engine implements image processing operators and an abstract image format.

Using these abstract classes allows the user to write image processing code that can run with various different image engines (eg imagej or opencv).
"""

import abc

ABC = abc.ABCMeta('ABC', (object,), {})


class IImage(ABC):

    @abc.abstractmethod
    def width(self):
        pass

    @abc.abstractmethod
    def height(self):
        pass


class IImageEngine(ABC):
    """
    The active image processor.

    As there should only be one image engine instance at any given time, this class is implemented as a singleton
    """

    # Here will be the instance stored.
    __instance = None

    @staticmethod
    def set_instance(image_engine):
        """Change the active image_engine.

        :param IImageEngine image_engine:
        """
        IImageEngine.__instance = image_engine

    @staticmethod
    def get_instance():
        """Static access method.
        :rtype IImageEngine:
        """

        assert IImageEngine.__instance is not None
        return IImageEngine.__instance

    @abc.abstractmethod
    def save_as_tiff(self, image, out_file_path):
        """
        :param IImage image:
        :param str out_file_path: eg './white_estimate.tiff'
        """

    @abc.abstractmethod
    def compute_max(self, images_file_path):
        """Compute for each pixel position the maximum at this position in all input images.

        :param list(str) images_file_path:
        :rtype IImage:
        """

    @abc.abstractmethod
    def mean_filter(self, image, radius):
        """Each pixel becomes an average of its neighbours within the given radius

        :param IImage image:
        :param int radius: in pixels
        """

    @abc.abstractmethod
    def perform_gray_morphology(self, image, operator, structuring_element_shape, structuring_element_radius):
        """
        :param IImage image:
        :param str operator: eg 'open'
        :param str structuring_element_shape: eg 'square'
        :param int structuring_element_radius:
        """

    @abc.abstractmethod
    def replace_border(self, image, band_width):
        """Overwrites the outer band of the image by duplicating the value of the pixels that touch this outer band

        :param IImage image:
        :param int band_width: width of the outer band, in pixels
        :rtype IImage:

        adaptation for the following code from matlab telemosToolbx's estimatedwhiteFluoImageTelemos function
            outer = white_estimate;
            white_estimate = white_estimate(4:(end-3),4:(end-3));
            outer(1:3,4:(end-3))=repmat(white_estimate(1,:),3,1);  # top
            outer((end-2):end,4:(end-3))=repmat(white_estimate(end,:),3,1); # bottom
            outer(:,1:3)=repmat(outer(:,4),1,3); # left
            outer(:,(end-2):end)=repmat(outer(:,end-3),1,3); # right
            white_estimate = outer;
            clear outer;
        """
        raise NotImplementedError()
        return image


