"""Image processing using imagej.
"""

from imageengine import IImage, IImageEngine
from ij import IJ  # pylint: disable=import-error
from ij.plugin import ImageCalculator  # pylint: disable=import-error
from ijopencv.ij import ImagePlusMatConverter  # pylint: disable=import-error
from ijopencv.opencv import MatImagePlusConverter  # pylint: disable=import-error
import org.bytedeco.javacpp.opencv_core as opencv_core  # pylint: disable=import-error
import org.bytedeco.javacpp.opencv_imgproc as opencv_imgproc  # pylint: disable=import-error
# from org.bytedeco.javacpp.opencv_imgcodecs import imread, imwrite  # pylint: disable=import-error


class IJImage(IImage):

    def __init__(self, image_engine, ij_image):
        """
        :param IJImageEngine image_engine:
        :param ImagePlus ij_image:
        """
        self.image_engine = image_engine
        self.ij_image = ij_image

    def width(self):
        raise NotImplementedError

    def height(self):
        raise NotImplementedError


def imagej_run_image_command(image, command, options):
    """performs the given imagej command on the given image

    :param ImagePlus image:
    :param str command: imagej command (eg "Gray Morphology")
    :param str options: imagej options (eg "radius=1 type=square operator=open")

    wrapper around IJ.run (https://imagej.nih.gov/ij/developer/api/ij/IJ.html#run-ij.ImagePlus-java.lang.String-java.lang.String-) which raises an exception on error
    """
    IJ.run(image, command, options)
    error_message = IJ.getErrorMessage()
    if error_message is not None:
        raise Exception('The command "%s" with options "%s" failed because of the following error : %s' % (command, options, error_message))


def perform_gray_morphology_with_imagej(image, operator, structuring_element_shape, structuring_element_radius):
    """
    :param IJImage image:
    :param str operator: eg 'open'
    :param str structuring_element_shape: eg 'square'
    :param int structuring_element_radius:
    """
    assert operator not in ['fast open', 'fast erode'], "as of 13/09/2019, fast operators such as 'fast erode' seem broken in fiji (the resulting image is not at all similar to their slow equivalent)"
    
    processor = image.ij_image.getProcessor()
    convert_to_byte = False
    if processor.getBitDepth() != 8:
        convert_to_byte = True
        min_value = processor.getMin()
        max_value = processor.getMax()
        print("warning: downgrading image to byte as imagej's Gray Morphology processing doesn't handle 16bit images (range=[%d, %d])" % (min_value, max_value))
        do_scaling = True
        image.ij_image.setProcessor(processor.convertToByte(do_scaling))
    print("before gray morphology")
    assert structuring_element_radius < 11, "the radius of the structuring element is too big (%d); using it with Fiji's 'Gray Morphology' tool would result in very long computations." % structuring_element_radius
    imagej_run_image_command(image.ij_image, "Gray Morphology", "radius=%d type=%s operator=%s" % (structuring_element_radius, structuring_element_shape, operator))
    print("after gray morphology")


def perform_gray_morphology_with_ijopencv(image, operator, structuring_element_shape, structuring_element_radius):
    """Perform gray morphology on the given image using imagej's opencv api (this api is provided by javacpp-presets)

    :param IJImage image:
    :param str operator: eg 'open'
    :param str structuring_element_shape: eg 'square'
    :param int structuring_element_radius:
    """

    shape_name_to_opencv_id = {
        'square': opencv_imgproc.MORPH_RECT,
    }

    operator_name_to_opencv_id = {
        'open': opencv_imgproc.MORPH_OPEN,
        'close': opencv_imgproc.MORPH_CLOSE,
    }

    if structuring_element_shape not in shape_name_to_opencv_id.keys():
        raise NotImplementedError('structuring element shape %s is not currently supported (supported operators : %s)' % (structuring_element_shape, str(operator_name_to_opencv_id.keys())))

    if operator not in operator_name_to_opencv_id.keys():
        raise NotImplementedError('operator %s is not currently supported (supported operators : %s)' % (operator, str(operator_name_to_opencv_id.keys())))

    imp2mat = ImagePlusMatConverter()
    cv_src_image = imp2mat.toMat(image.ij_image.getProcessor())

    # cv_src_image = opencv_core.Mat().zeros(128, 128, opencv_core.CV_8U).asMat()
    print(cv_src_image)

    struct_el_size = structuring_element_radius * 2 + 1

    struct_el_size = opencv_core.Size(struct_el_size, struct_el_size)
    struct_el_anchor = opencv_core.Point(structuring_element_radius, structuring_element_radius)
    # print(struct_el_anchor)
    struct_element = opencv_imgproc.getStructuringElement(shape_name_to_opencv_id[structuring_element_shape], struct_el_size, struct_el_anchor)
    print(struct_element)
    # dst_image = opencv_core.Mat().zeros(128, 128, opencv_core.CV_8U).asMat()
    cv_dst_image = opencv_core.Mat(cv_src_image.size(), cv_src_image.type())

    # warning ! trying to guess the arguments from opencv's documentation is time consuming as the error messages are misleading (in the following call, javacpp complains that the 1st argument is not a org.bytedeco.javacpp.opencv_core$Mat, while it is ! The problem comes from the order of the arguments). So, instead of guessing, the found accepted signatures of morphologyEx are in javacpp-presets/opencv/src/gen/java/org/bytedeco/opencv/global/opencv_imgproc.java
    # opencv_imgproc.morphologyEx(cv_src_image, opencv_imgproc.MORPH_OPEN, struct_element, dst_image)
    # TypeError: morphologyEx(): 1st arg can't be coerced to org.bytedeco.javacpp.opencv_core$GpuMat, org.bytedeco.javacpp.opencv_core$Mat, org.bytedeco.javacpp.opencv_core$UMat
    print('before opencv_imgproc.morphologyEx')
    opencv_imgproc.morphologyEx(cv_src_image, cv_dst_image, operator_name_to_opencv_id[operator], struct_element)
    print('after opencv_imgproc.morphologyEx')

    mat2imp = MatImagePlusConverter()
    image.ij_image.setProcessor(mat2imp.toImageProcessor(cv_dst_image))


class IJImageEngine(IImageEngine):

    def missing(self):
        pass

    def save_as_tiff(self, image, out_file_path):
        IJ.saveAsTiff(image.ij_image, out_file_path)

    def compute_max(self, images_file_path):
        """Computes for each pixel position the maximum at this position in all input images.

        :param list(str) images_file_path:
        :rtype IJImage:
        """
        assert len(images_file_path) > 1
        max_image = IJ.openImage(images_file_path[0])
        print('max_image', max_image)

        for image_file_path in images_file_path[2:-1]:
            other_image = IJ.openImage(image_file_path)
            print('other_image', other_image)
            ic = ImageCalculator()
            ic.run("max", max_image, other_image)
        print('max_image', max_image)
        return IJImage(self, max_image)

    def mean_filter(self, image, radius):
        """Implement interface method."""
        IJ.run(image.ij_image, "Mean...", "radius=%d" % radius)

    def perform_gray_morphology(self, image, operator, structuring_element_shape, structuring_element_radius):
        """
        :param IJImage image:
        :param str operator: eg 'open'
        :param str structuring_element_shape: eg 'square'
        :param int structuring_element_radius:
        """
        # we use opencv version because imagej's slow version is way too slow for big values of structuring_element_radius
        perform_gray_morphology_with_ijopencv(image, operator, structuring_element_shape, structuring_element_radius)

    def replace_border(self, image, band_width):
        """Implement interface method."""
        raise NotImplementedError()
        return image



